/***
 * Excerpted from "The ThoughtWorks Anthology 2",
 * published by The Pragmatic Bookshelf.
 * Copyrights apply to this code. It may not be used to create training material, 
 * courses, books, articles, and the like. Contact us if you are in doubt.
 * We make no guarantees that this code is fit for any purpose. 
 * Visit http://www.pragmaticprogrammer.com/titles/twa2 for more book information.
***/
package voterlist.classbased;

import java.util.Collection;
import java.util.HashSet;

import static java.util.Arrays.asList;

class CouncilOfNations {
    private Collection<Country> memberNations;
    public CouncilOfNations(Collection<Country> memberNations) {
        this.memberNations = memberNations;
    }
    public boolean contains(Country country) {
        return memberNations.contains(country);
    }
}

class Country {
    private String name;
    public Country(String name) {
        this.name = name;
    }
    public String toString() {
        return name;
    }
}

class Person {
    private String name;
    private int age;
    private Country country;
    public Person(String name, int age, Country country) {
        // ...
        this.name = name;
        this.age = age;
        this.country = country;
    }
    public boolean canVoteIn(Country votingJurisdiction) {
        return age >= 18 && votingJurisdiction.equals(country);
    }
    public String toString() {
        return name;
    }
}

abstract class AbstractVoterList<T, X> {
    private Collection<T> candidateVoters;
    public AbstractVoterList(Collection<T> candidateVoters) {
        this.candidateVoters = candidateVoters;
    }
    public Collection<T> votersFor(X votingJurisdiction) {
        Collection<T> eligibleVoters = new HashSet<T>();
        for (T voter : candidateVoters) {
            if (canVoteIn(voter, votingJurisdiction)) {
                eligibleVoters.add(voter);
            }
        }
        return eligibleVoters;
    }
    protected abstract boolean canVoteIn(T voter, X votingJurisdiction);
}
class PersonVoterList extends AbstractVoterList<Person, Country> {
    public PersonVoterList(Collection<Person> persons) {
        super(persons);
    }
    protected boolean canVoteIn(Person person, Country country) {
        return person.canVoteIn(country);
    }
}
class CountryVoterList extends AbstractVoterList<Country, CouncilOfNations> {
    public CountryVoterList(Collection<Country> countries) {
        super(countries);
    }
    protected boolean canVoteIn(Country country,
                                CouncilOfNations councilOfNations) {
        return councilOfNations.contains(country);
    }
}

public class VoterListClassBased {
    public static void main(String[] args) {
        Country INDIA = new Country("India");
        Country USA = new Country("USA");
        Country UK = new Country("UK");
        Collection<Person> persons = asList(
            new Person("Donald", 28, INDIA),
            new Person("Daisy", 25, USA),
            new Person("Minnie", 17, UK)
        );
        PersonVoterList personVoterList = new PersonVoterList(persons);
        System.out.println(personVoterList.votersFor(INDIA)); // [ Donald ]
        System.out.println(personVoterList.votersFor(USA));   // [ Daisy ]
        Collection<Country> countries = asList(INDIA, USA, UK);
        CountryVoterList countryVoterList = new CountryVoterList(countries);
        CouncilOfNations councilOfNations = new CouncilOfNations(asList(
            USA, INDIA
        ));
        System.out.println(countryVoterList.votersFor(councilOfNations));
        // [ USA, India ]
    }
}
